/* lqkwic.c -- Copyright 1991, 1992 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* lqkwic -- produce a keyword-in-context list of matches...
 * Liam R. Quin, February 1991 and later...
 *
 * $Id: lqkwic.c,v 1.26 1993/12/09 04:19:34 lee Exp $
 */

/* lqkwic presents user with a keyword-in-context (KWIC) index,
 * by looking up each match in the file, getting the surrounding
 * text, and displaying it.
 *
 * lqkwic can generate various formatted output:
 *   + text before the matches for a new file-name
 *   + text before each match
 *   + text after each match
 *   + text between the components of a match
 *   + specified amounts of context before and after the match
 *   + specified portions of the file name
 *
 */


#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#ifdef CURSESX
# include <curses.h>
# include <term.h>
#endif

#include <malloc.h>
#include <ctype.h>

#include <sys/types.h> /* for fileinfo.h */

#ifndef FILE
# include <stdio.h>
#endif

#include "range.h"
#include "fileinfo.h"
#include "wordinfo.h"
#include "wordrules.h"
#include "emalloc.h"
#include "readfile.h"
#include "wmoffset.h"

/** Unix system calls that need declaring: **/
extern long lseek();
extern int open(), close();
extern int read();
extern void exit();

/** Unix/C Library Functions that need declaring: **/
#ifndef tolower
 extern int tolower();
#endif
extern int strlen();
extern int atoi();
extern long atol();

/** lqtext library functions that need declaring: **/
extern void SetDefaults();
extern void DefaultUsage();

/** Functions within this file that are used before being defined: **/
static char *GetVariableValue();
int ShowOneMatch();
static char *EndsEntity();
static char *StartsEntity();
static char *EntityValue();
static void ReadEntityFile();
static void ListVariables();
static int StringRefersToVariable();
extern char *DownCase();

/** **/

/** **/

extern int errno;

int AsciiTrace = 0;
char *progname = "lqkwic"; /* set from argv[] in main() */

static char *Revision = "@(#) $Id: lqkwic.c,v 1.26 1993/12/09 04:19:34 lee Exp $";

static int TruncateAtPath = 0;
static int MatchesHaveWordCount = 1;
#define MATCHLEN (3 + MatchesHaveWordCount)

/* Default string segment widths:
 * Use the screen width for the overall amount, dividing the
 * remaining space evenly between left and right:
 */
#define DEFAULT_SCREENWIDTH 75 /* in case we can't deduce it */

/* Default buffer sizes: */
static int LeftWidth = DEFAULT_SCREENWIDTH * 2;
static int RightWidth = DEFAULT_SCREENWIDTH * 2;
static int GapWidth = 0;

static char ShowAllLines = 0;
static unsigned long LineNumber = 0L;
static char *MatchFileFormat =
	"==== Document ${FileNumber}: ${FileName} ====${NEWLINE}";
static char *MatchFormat =
"$[${MatchNumber/3r}:${TextBefore/30r 30l}${Gap}${MatchedText}${TextAfter}/79 79r]${NEWLINE}";

static int SuppressTags = 0; /* suppress SGML tags */
static char TagMark = '/';

static int NiceWarnIfEmpty = 0;

static t_Range *RangeOfMatchesToPrint = 0;
static t_Range *RangeOfFilesToPrint = 0;

/* characters used for variable expansion: */
#define BRA '{'
#define KET '}'

/* lazy evalyation flags */
typedef struct {
    int NeedText;
    int NeedFileName;
} t_LazyEvaluation;

static t_LazyEvaluation LazyEvaluation = {
    0, 0
};

int
main(argc, argv)
    int argc;
    char *argv[];
{
    extern char *getenv();
    extern int optind, getopt();
    extern char *optarg; /* for getopt */
    int ch; /* for getopt */
    int ErrFlag = 0; /* see how getopt makes programs cleaner? */
    char *FileWithMatches = (char *) 0;
    int SeenMatches = 0;
    int NeedVariableList = 0;

    {
	char *p = getenv("COLS");
	if (p && *p) {
	    if (!cknatstr(p)) {
		Error(E_WARN,
		    "Environent variable $COLS has non-numeric value \"%s\"",
		    p
		);
	    } else {
		int Width;

		Width = atoi(p);
		if (Width > 0 && Width != DEFAULT_SCREENWIDTH) {
		    LeftWidth = Width / 2;
		    RightWidth = (Width + 1) / 2;
		}
	    }
	}
    }

    progname = argv[0];

    SetDefaults(argc, argv);

    /* All lq-text programs must call SetDefaults() before getopt, and
     * must then be prepared to ignore options z with arg and Z without.
     */
    while ((ch = getopt(argc, argv, "AC:cE:ef:g:Ll:Mo:O:p:r:S:s:Vvxz:Z")) != EOF) {
	switch (ch) {
	case 'Z':
	case 'z':
	    break; /* done by SetDefaults(); */
	case 'V':
	    fprintf(stderr, "%s version %s\n", progname, Revision);
	    break;
	case 'v':
	    AsciiTrace = 1;
	    break;
	case 'A':
	    ShowAllLines = 1;
	    break;
	case 'c':
	    SuppressTags = 1;
	    break;
	case 'C':
	    TagMark = optarg[0];
	    break;
	case 'e':
	    NiceWarnIfEmpty = 1;
	    break;
	case 'g':
	    if (!cknatstr(optarg)) {
		Error(E_FATAL|E_USAGE|E_XHINT,
		    "-g must be followed by a whole number, not \"%s\"",
		    optarg
		);
	    }
	    GapWidth = atoi(optarg);
	    break;
	case 'E':
	    ReadEntityFile(optarg);
	    break;
	case 'f':
	    FileWithMatches = optarg;
	    break;
	case 'l':
	    LeftWidth = atoi(optarg);
	    break;
	case 'L':
	    NeedVariableList = 1;
	    break;
	case 'M':
	    MatchesHaveWordCount = 0;
	    break;
	case 'o':
	    RangeOfMatchesToPrint = StringToRange(optarg);
	    break;
	case 'O':
	    RangeOfFilesToPrint = StringToRange(optarg);
	    break;
	case 's':
	    MatchFormat = optarg;
	    break;
	case 'p':
	    TruncateAtPath = atoi(optarg);
	    if (!TruncateAtPath) {
		Error(E_FATAL|E_XHINT|E_USAGE,
		    "-p: must be followed by number of / characters to delete"
		);
	    }
	    break;
	case 'r':
	    RightWidth = atoi(optarg);
	    break;
	case 'S':
	    MatchFileFormat = optarg;
	    break;
	case 'x':
	    ErrFlag = (-1);
	    break;
	default:
	    Error(E_WARN,
		"option -%c is not recognised by this program",
		ch
	    );
	case '?':
	    ErrFlag = 1;
	}
    }

    if (ErrFlag < 0) { /* -x or -xv was used */
	fprintf(stderr, "usage: %s [-xv] [options] [matches...]\n", progname);
	fprintf(stderr,
	"use %s -x, -xv or -xvv for more detailed explanations.\n", progname);

	if (AsciiTrace) {
	    fprintf(stderr, "\n\
	-e      -- an empty match file indicates no matches were found\n\
	-E file -- \"file\" contains SGML entity names and values\n");
	    if (AsciiTrace > 1) {
		fprintf(stderr, "\
                   The format of the entity file is:\n\
# blank lines and comments are ignored;\n\
# comments start with a # and go on the end of the line.\n\
		   name \"replacement value\"\n\
# name is the name of an entity.  Case is significant.\n\
# the replacement value must be quoted if it contains\n\
# any spaces; you can use \"...\", `...' or '...'\n\
# Entities must be marked as &name; in the text.");
	    }
	    fprintf(stderr, "\
	-f file -- \"file\" contains a list of matches, one per line\n\
	-l n    -- display n characters to the left of each phrase [%d]\n\
	-o range - print only matches falling within the given range\n\
	-O range - print only matches in documents within the given range\n",
		    LeftWidth
	    );
	    if (AsciiTrace > 1) {
		fprintf(stderr, "\
                   The format of a range is:\n\
                   -12,14-16,130-200,301,400 417 796,1003,1800-\n\
		   where a leading -12 means to print everything up to and\n\
		   including the twelth item;\n\
		   a trailing 1800- means to print item 1800 and following;\n\
		   14-27 means to print items 14, 15 and 16;\n\
		   the other numbers standing for themselves;\n\
		   commas (,) and spaces are interchangeable at pleasure.\n\n"
		);
	    }
	    fprintf(stderr, "\
	-r n    -- display r chars to the right of each phrase's start [%d]\n\
	-p n    -- truncate pathnames at the nth rightmost `/' [%d]\n",
		    RightWidth,
		    TruncateAtPath
	    );
	}

	if (AsciiTrace > 1) {
	    fputs("\
	Matches should be in the form of\n\
		NumberOfWordsInPhrase  BlockNumber  WordInBlock  FileName\n\
	(This format is produced by lqrank, lqword -l, and lqphrase)\n\
", stderr);
	}
	DefaultUsage();

	if (NeedVariableList) {
	    ListVariables();
	}
	exit(0);
    } else if (ErrFlag > 0) {
	fprintf(stderr, "use %s -x for an explanation.\n", progname);
	exit(1);
    }

    if (NeedVariableList) {
	ListVariables();
	exit(0);
    }

    /* lazy evaluation */
    if (MatchFileFormat && *MatchFileFormat) {
	LazyEvaluation.NeedText = (
	    /* these are in guessed order of likelihood...
	     * there's another list like this on the next screenm, watch out!
	     */
	    StringRefersToVariable(MatchFileFormat, "TextBefore") ||
	    StringRefersToVariable(MatchFileFormat, "TextAfter") ||
	    StringRefersToVariable(MatchFileFormat, "MatchedText") ||
	    StringRefersToVariable(MatchFileFormat, "StartByte") ||
	    StringRefersToVariable(MatchFileFormat, "EndByte") ||
	    StringRefersToVariable(MatchFileFormat, "LeftPad")
	);
	LazyEvaluation.NeedFileName = (
	    StringRefersToVariable(MatchFileFormat, "FileName")
	);

    }

    if (MatchFormat && *MatchFormat) {
	if (!LazyEvaluation.NeedText ) {
	    LazyEvaluation.NeedText = (
		StringRefersToVariable(MatchFormat, "TextBefore") ||
		StringRefersToVariable(MatchFormat, "TextAfter") ||
		StringRefersToVariable(MatchFormat, "MatchedText") ||
		StringRefersToVariable(MatchFormat, "StartByte") ||
		StringRefersToVariable(MatchFormat, "EndByte") ||
		StringRefersToVariable(MatchFormat, "LeftPad")
	    );
	}
	if (!LazyEvaluation.NeedFileName) {
	    LazyEvaluation.NeedFileName = (
		StringRefersToVariable(MatchFormat, "FileName")
	    );
	}
    }

    /* check that we can get at the file containing the matches, if one
     * was supplied.
     */
    if (FileWithMatches) {
	if (*FileWithMatches == '-' && FileWithMatches[1] == '\0') {
	    if (fReadMatchFile(stdin, "standard input") < 0) {
		Error(E_FATAL|E_SYS,
		    "couldn't read matches from standard input"
		);
	    }
	} else {
	    extern FILE *fEopen();

	    FILE *f = fEopen(
		E_FATAL,
		FileWithMatches,
		"list of matches",
		"r"
	    );
	    /* Now read the file, and make an array of matches... */
	    if (fReadMatchFile(f, FileWithMatches) < 0) {
		Error(E_FATAL|E_SYS,
		    "couldn't read matches from %s",
		    FileWithMatches
		);
	    }
	    (void) fclose(f);
	    ++SeenMatches;
	}
    }

    argv += optind;
    argc -= optind;

    if (SeenMatches) {
	if (argc > 0) {
	    Error(E_FATAL|E_USAGE|E_XHINT,
		"unknown argument %s",
		*argv
	    );
	}
    } else {
	if (fReadMatchFile(stdin, "Standard input") < 0) {
	    Error(E_FATAL|E_SYS,
		"couldn't read matches from standard input"
	    );
	}
    }

    return 0;
}

long
fReadMatchFile(FileWithMatches, MatchFileName)
    FILE *FileWithMatches;
    char *MatchFileName;
{
    long MatchCount = 0;
    char *Line;
    int i;
    int NumberOfWords = 1;
    long BlockInFile;
    char *FileName;
    int WordInBlock;
    t_FID FID;

    if (!FileWithMatches) {
	Error(E_FATAL, "match-list file (from -f -) has NULL file!");
    }

    while (fReadLine(
	FileWithMatches,
	&Line,
	UF_IGNBLANKS|UF_IGNSPACES|UF_IGNHASH|UF_ESCAPEOK 
    ) >= 0) {
	register char *p;

	p = Line;

	/* ASSERT: There are no leading or trailing spaces on the line */
	if (!*p || *p == '#') {
	    (void) efree(Line);
	    continue; /* blank line */
	}

	if (MatchesHaveWordCount) {
	    if (!isdigit(*p)) {
		Error(E_WARN,
		    "%s: Bad match format (expected digit) in %s",
		    MatchFileName,
		    Line
		);
		(void) efree(Line);
		return MatchCount;
	    }

	    NumberOfWords = 0;

	    while (isdigit(*p)) {
		NumberOfWords *= 10;
		NumberOfWords += *p++ - '0';
	    }

	    while (isspace(*p)) {
		p++;
	    }
	}

	/* block in file */

	BlockInFile = 0L;

	while (isdigit(*p)) {
	    BlockInFile *= 10;
	    BlockInFile += *p - '0';
		p++;
	}

	while (isspace(*p)) {
	    p++;
	}

	/* Word In Block */

	WordInBlock = 0L;

	while (isdigit(*p)) {
	    WordInBlock *= 10;
	    WordInBlock += *p - '0';
		p++;
	}

	while (isspace(*p)) {
	    p++;
	}

	/* file identifier (FID) */
	FID = 0L;

	while (isdigit(*p)) {
	    FID *= 10;
	    FID += *p - '0';
		p++;
	}

	while (isspace(*p)) {
	    p++;
	}

	/* filename */

	if (!*p || !FID) {
	    static char *theName = 0;
	    static t_FID theFID = 0L;
	    t_FileInfo *FileInfo;

	    if (!FID) {
		Error(E_FATAL,
		    "%s: bad match format, neither FID nor filename",
		    MatchFileName
		);
	    }
	    
	    if (!FID) {
		if (theName && STREQ(p, theName)) {
		    FID = theFID;
		} else {
		    FID = Name2FID(FileName);
		}
	    }

	    if (!*p) {
		if (FID == theFID) {
		    FileName = theName;
		} else {
		    extern t_FileInfo *GetFileInfo();

		    if (theName) {
			efree(theName);
		    }
		    FileInfo = GetFileInfo(FID);
		    theFID = FID;
		    theName = FileName = FileInfo->Name;
		    (void) efree(FileInfo);
		}
	    }
	} else {
	    /* file name, already null-terminated */
	    FileName = p;
	}

	++MatchCount;
	(void) ShowOneMatch(
	    MatchCount,
	    NumberOfWords,
	    BlockInFile,
	    WordInBlock,
	    FileName,
	    FID
	);
	efree(Line);
    }

    return MatchCount;
}

int
ShowOneMatch(
    MatchNumber,
    WordCount,
    BlockInFile,
    WordInBlock,
    DocumentName,
    FID
)
    long MatchNumber;
    char *DocumentName;
    unsigned long BlockInFile;
    unsigned int WordInBlock;
    int WordCount;
    t_FID FID;
{
    static char *Buffer = 0;
    static char *LastDocumentName = 0;
    static char *LastFileName = 0;
    static unsigned long LastBlockInFile = (unsigned long) -1L;
    static int AmountRead;
    static int fd = -1;

    register char *p;
    register char *q;

    int InSpace = 0;
    t_OffsetPair *OffsetPair = 0;
    char *FirstStart = 0;
    char *TextBefore = 0;
    char *MatchedText = 0;
    char *TextAfter = 0;
    char *FileName = DocumentName;

    /** see if the file name has changed **/
    if (LastDocumentName == (char *) 0 ||
			    !STREQ(LastDocumentName, DocumentName)) {
	if (LastDocumentName != (char *) 0) {
	    if (fd >= 0) {
		(void) close(fd);
	    }
	    efree(LastDocumentName);
	}
	fd = (-1);

	LastDocumentName = emalloc("LastDocmnt", strlen(DocumentName) + 1);
	(void) strcpy(LastDocumentName, DocumentName);
	LastBlockInFile = (unsigned long) (unsigned long) -1L;
	if (LastFileName) {
	    (void) efree(LastFileName);
	}
	FileName = FindFile(DocumentName);
	if (!FileName) {
	    Error(E_WARN|E_SYS, "can't find file \"%s\"", DocumentName);
	    return -1;
	}
	LastFileName = emalloc("LastFileName", strlen(FileName) + 1);
	(void) strcpy(LastFileName, FileName);

    }

    /** if necessary, open the file and read the text **/

    if (LazyEvaluation.NeedText) {
	static unsigned int BufLen;

	if (fd < 0) {
	    errno = 0;

#ifdef COMPRESS_SUPPORT
	    fd = UnpackAndOpen(FileName);
#else /* COMPRESS_SUPPORT */
	    fd = open(FileName, O_RDONLY, 0);
#endif /* COMPRESS_SUPPORT */

	    if (fd < 0) {
		return -1;
	    }
	}

	if (LastBlockInFile != BlockInFile) {
	    if (lseek(fd,
		BlockInFile ? (long) ((BlockInFile - 1) * FileBlockSize) : 0L,
		0
	    ) < 0) {
		Error(E_SYS|E_WARN, "%s: lseek(%ld) system call failed",
		    DocumentName,
		    BlockInFile ?
			(long) ((BlockInFile - 1) * FileBlockSize)
		    :
			0L,
		    0
		);
		return -1;
	    }

	    /** Ensure that we have a line buffer, if we will need one **/
	    if (Buffer == (char *) 0) {

		/* TODO: use the real width from the variable formats */
		BufLen = (LeftWidth + RightWidth) * 5;
		/* Make enough space for a block before and a block after: */
		if (BufLen < FileBlockSize * 3) {
		    BufLen = FileBlockSize * 3;
		}
		Buffer = emalloc("lqkwic line buffer", BufLen);
	    }

	    if ((AmountRead = read(fd, Buffer, BufLen)) < MinWordLength) {
		Error(E_SYS|E_WARN, "%s: read(%d,0x%x,%d) returned %d",
		    DocumentName,
		    fd, Buffer, BufLen,
		    AmountRead
		);
		(void) close(fd);
		return -1;
	    }
	    LastBlockInFile = BlockInFile;
	}

	/** Find the required word */

	OffsetPair = FindMatchEnds(
	    Buffer,
	    AmountRead,
	    (BlockInFile == 0) ?
		Buffer : &Buffer[FileBlockSize],
	    BlockInFile, /* why is this needed?? */
	    WordInBlock,
	    WordCount
	);

	if (!OffsetPair) {
	    Error(E_WARN, "Match %d (block %ld, word %d, in %s) not found",
		LineNumber,
		BlockInFile,
		WordInBlock,
		DocumentName
	    );
	    return 0; /* but carry on */
	}

	/* Find context before the keyword */

	TextBefore = emalloc("lqkwic match.firstbit", LeftWidth + 1);
	q = &TextBefore[LeftWidth];
	*q = '\0';

	InSpace = 0;
	{
	    int WithinTag = 0;
	    int TagMarkInserted = 0;

	    for (p = OffsetPair->Start - 1; p >= Buffer; --p) {
		char *Entity;

		if (q == TextBefore) {
		    break;
		}
		if (SuppressTags) {
		    if (WithinTag) {
			if (*p == '<') {
			    WithinTag = 0;
			}
			continue;
		    } else if (*p == '>') {
			if (!TagMarkInserted) {
			    *--q = TagMark;
			    TagMarkInserted = 1;
			}
			WithinTag = 1;
			continue;
		    } else if ((Entity = EndsEntity(p, Buffer)) != 0) {
			char *Value = EntityValue(Entity);

			/* Insert the entity value backwards into the text */
			if (Value && *Value) {
			    char *z = Value;

			    /* find the end: */
			    while (*z) z++;

			    while (z > Value && q > TextBefore) {
				*--q = *--z;
			    }
			    /* skip to the start of the entity: */
			    while (*p != '&') {
				p--;
			    }
			    (void) efree(Entity);
			    TagMarkInserted = 0;
			    continue;
			} else {
			    (void) efree(Entity);
			    TagMarkInserted = 0;
			}
		    } else if (!isspace(*p)) {
			TagMarkInserted = 0;
		    }
		}

		if (isspace(*p)) {
		    if (!InSpace) {
			*--q = ' ';
		    }
		    InSpace = 1;
		} else {
		    InSpace = 0;
		    *--q = *p;
		}
	    }
	}

	FirstStart = q;

	/* now build up the rest of the buffer */

	q = MatchedText = emalloc("lqkwic MatchedText", RightWidth + 1);
	*q = '\0';
	TextAfter = emalloc("lqkwic TextAfter", RightWidth + 1);
	*TextAfter = '\0';

	InSpace = 0;

	{
	    int WithinTag = 0;
	    int TagMarkInserted = 0;
	    int Length = 0;

	    for (p = OffsetPair->Start; p - Buffer < AmountRead; p++) {

		char *Entity;

		if (++Length > RightWidth) break;

		if (p == OffsetPair->End) {
		    *q = '\0';
		    q = TextAfter;
		}

		if (SuppressTags) {
		    if (WithinTag) {
			if (*p == '>') {
			    WithinTag = 0;
			}
			continue;
		    } else if (*p == '<') {
			if (!TagMarkInserted) {
			    *q++ = TagMark;
			    TagMarkInserted = 1;
			}
			WithinTag = 1;
			continue;
		    } else if ((Entity = StartsEntity(p)) != 0) {
			char *Value = EntityValue(Entity);
			/* Insert the entity value into the text */
			if (Value && *Value) {
			    while (q < TextAfter && *Value) {
				*q++ = *Value++;
			    }
			    /* skip to the end of the entity: */
			    while (*p && *p != ';') {
				p++;
			    }
			    TagMarkInserted = 0;
			    (void) efree(Entity);
			    InSpace = 0;
			    continue;
			} else {
			    TagMarkInserted = 0;
			    (void) efree(Entity);
			    InSpace = 0;
			    /* but no continue; */
			}
		    } else if (!isspace(*p)) {
			TagMarkInserted = 0;
		    }
		}

		if (isspace(*p)) {
		    if (!InSpace) {
			*q = ' ';
			*++q = '\0';
		    }
		    InSpace = 1;
		} else {
		    *q = *p;
		    *++q = '\0';
		    InSpace = 0;
		}
	    }
	}
    }

    {
	unsigned long StartByte, EndByte;

	if (OffsetPair) {
	    StartByte = (OffsetPair->Start - Buffer) + 
		(BlockInFile ? (long) ((BlockInFile - 1) * FileBlockSize) : 0L);
	    EndByte = StartByte + (OffsetPair->End - OffsetPair->Start);
	} else {
	    /* in this case they won't be used anyway, because we checked
	     * for LazyEvaluation earlier.
	     */
	    StartByte = EndByte = 0L;
	}

	OutputMatch(
	    MatchFileFormat, MatchFormat,
	    MatchNumber, FID, LastDocumentName, LastFileName,
	    WordCount, BlockInFile, WordInBlock,
	    FirstStart, MatchedText, TextAfter,
	    StartByte, EndByte
	);
    }

    if (LazyEvaluation.NeedText) {
	efree(TextBefore);
	efree(TextAfter);
	efree(MatchedText);
    }
    return 0;
}

typedef struct s_Entity {
    char *Name;
    char *Value;
    struct s_Entity *Next;
} t_Entity;

static t_Entity *EntityList = NULL;

static void
InsertEntity(e)
    t_Entity *e;
{
    t_Entity **Epp = &EntityList;

    /* Locate: */
    for (Epp = &EntityList; *Epp; Epp = &(*Epp)->Next) {
	if (STRCMP(e->Name, (*Epp)->Name) >= 0) {
	    break;
	}
    }

    /* Insert: */
    e->Next = (*Epp);
    *Epp = e;
#ifdef ASCIITRACE
    if (AsciiTrace > 1) {
	fprintf(stderr, "Insert Entity &%s; as \"%s\"\n", e->Name, e->Value);
    }
#endif
}

static
char *
GetNextString(Source, pp)
    char *Source; /* e.g. the file name */
    char **pp;
{
    char quotes = 0;
    char *Start;
    char *String;
    char *Result;

    if (!pp || !*pp || !**pp) return 0;

    String = (*pp); /* for error reporting */

    while (isspace(**pp)) ++*pp;

    switch (**pp) {
    case '\0':
	return (char *) NULL;
    case '"': case '\'':
	quotes = (**pp);
	++*pp;
	break;
    case '`':
	quotes = '\''; /* `...' */
	++*pp;
	break;
    default:
	break;
    }

    Start = (*pp);

    /* find the end... */
    while (**pp) {
	if ((quotes && **pp == quotes) || (!quotes && isspace(**pp))) {
	    break;
	}
	++*pp;
    }

    if (quotes) {
	if (**pp != quotes) {
	    Error(E_FATAL,
		"%s: unmatched quote in %s, expected %c",
		Source,
		String,
		quotes
	    );
	}

    }
    Result = emalloc("lqkwic::GetNextString.Result", *pp - Start + 2);
    (void) strncpy(Result, Start, *pp - Start + 1);
    Result[*pp - Start] = '\0';


    if (**pp) {
	++*pp;
    }

    return Result;
}

static char *
StartsEntity(String)
    char *String;
{
    register char *p;

    if (*String != '&') {
	return 0;
    }

    ++String; /* skip over the initial & */

    for (p = String; *p; p++) {
	if (isspace(*p)) return 0;
	if (*p == ';') {
	    char *Name = emalloc(
		"lqkwic:SGML Entity Name",
		(unsigned int)(p - String + 1)
	    );
	    (void) strncpy(Name, String, p - String);
	    Name[p - String] = '\0';
	    return Name;
	}
    }
    return 0;
}

static char *
EndsEntity(String, Buffer)
    char *String;
    char *Buffer;
{
    register char *p;

    if (*String != ';') {
	return 0;
    }

    for (p = String; p > Buffer; --p) {
	if (isspace(*p)) return 0;
	if (*p == '&') {
	    char *Name;
	    ++p; /* skip over the & */
	    Name = emalloc("lqkwic:EndsEntity.Name", String - p + 1);
	    (void) strncpy(Name, p, String - p);
	    Name[String - p] = '\0';
	    return Name;
	}
    }
    return 0;
}

static t_Entity *
NewEntity(String, FileName)
    char *String;
    char *FileName;
{
    t_Entity *e;

    e = (t_Entity *) emalloc("lqkwic:NewEntity.Name", sizeof (t_Entity));
    e->Next = 0;
    e->Name = GetNextString(FileName, &String);
    e->Value = GetNextString(FileName, &String);
    return e;
}

static void
ReadEntityFile(FileName)
    char *FileName;
{
    char **Lines = 0;
    int count;
    int i;
    
    count = ReadFile(
	E_FATAL,
	FileName,
	"SGML Entity Replacement File",
	&Lines, /* yes, a (char ***) */
	UF_IGNBLANKS|UF_IGNSPACES|UF_IGNHASH|UF_ESCAPEOK
    );

    for (i = 0; i < count; i++) {
	t_Entity *e = NewEntity(Lines[i], FileName);

	if (e) {
	    InsertEntity(e);
	}
	(void) efree(Lines[i]);
    }

    (void) efree((char *)Lines);
}

static char *
EntityValue(Name)
    char *Name;
{
    register t_Entity *Ep;
    int i = 1; /* i.e. non-zero in case loop is never reached */

    for (Ep = EntityList; Ep; Ep = Ep->Next) {
	if ((i = STRCMP(Name, Ep->Name)) >= 0) {
	    break;
	}
    }

    if (i == 0 && Ep) {
#ifdef ASCIITRACE
	if (AsciiTrace > 3) {
	    fprintf(stderr, "Entity &%s; has value \"%s\"\n", Name, Ep->Value);
	}
#endif
	return Ep->Value;
    } else {
	if (AsciiTrace > 3) {
#ifdef ASCIITRACE
	    fprintf(stderr, "Entity &%s; has no declared value\n", Name);
#endif
	}
	return 0;
    }
}

static char *FormatStringValue();

static char *
NumberToString(StringFormat, VariableName, Number, Format)
    char *StringFormat;
    char *VariableName;
    unsigned long Number;
    char *Format;
{
    static char NumberBuffer[30];

    (void) sprintf(NumberBuffer, Format, Number);
    return FormatStringValue(StringFormat, VariableName, NumberBuffer);
}

static int
StringRefersToVariable(String, Variable)
    char *String;
    char *Variable;
{
    char *theName = 0;
    register char *p;
    char First = Variable[0];

    /* We store the first character of the variable separately for
     * convenience...:
     */
    if (isupper(First)) {
	First = tolower(First);
    }

    ++Variable;
    
    for (p = String; *p; p++) {
	if (*p == '$' && p[1] == BRA &&
		(p[2] == First || (isupper(p[2]) && tolower(p[2]) == First))) {

	    register char *q, *np;

	    if (!theName) {
		/* only do this if we need to...
		 * it sometimes calls malloc() and can be slow:
		 */
		theName = DownCase(Variable);
	    }

	    ++p; /* step over the $ */
	    ++p; /* step over the { */
	    ++p; /* step over the first char of the name, already checked */

	    for (q = p, np = theName; *q; q++, np++) {
		if (*q == KET) {
		    return 1;
		} else if (*q != *np && !(isupper(*q) && *np == tolower(*q))) {
		    break;
		}
	    }

	    if (*q) {
		p = q;
	    } else {
		return 0;
	    }
	}
    }
    return 0;
}

/* This routine is due to change -- it's crazy to put this
 * much on the stack for each call!
 *
 */

static char *
sPrintWithFormat(Format,
    MatchFileFormat, MatchFormat,
    FileNumber, MatchNumber, MatchWithinFile,
    FID, DocName, FileName, LastDocName,
    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
    TextBefore, MatchedText, TextAfter,
    StartByte, EndByte
)
    char *Format;
    char *MatchFileFormat;
    unsigned long FileNumber;
    long MatchNumber;
    unsigned long MatchWithinFile;
    char *MatchFormat;
    t_FID FID;
    char *DocName; /* the name stored in the database */
    char *FileName; /* the expanded file name */
    char *LastDocName; /* the name stored in the database */
    int NumberOfWordsInPhrase;
    unsigned long BlockInFile;
    int WordInBlock;
    char *TextBefore;
    char *MatchedText;
    char *TextAfter;
    unsigned long StartByte, EndByte;
{
    register char *p;
    char *Result;
    unsigned int AllocatedLength = 100;
    char *Rp;

    Result = emalloc("sPrintWithFormat", AllocatedLength);
    Rp = Result;
    *Rp = '\0';

#define ADDCHAR(ch) \
    if (Rp - Result > AllocatedLength - 2) { \
	int Where = Rp - Result; \
	Result = erealloc(Result, AllocatedLength += 50); \
	Rp = &Result[Where]; \
    } \
    *Rp = (ch); \
    *++Rp = '\0';

#define ADDSTRING(str) \
    { \
	register char *_p; \
 \
	for (_p = str; *_p; _p++) { \
	    ADDCHAR(*_p); \
	} \
    } \

    for (p = Format; *p; p++) {
	if (*p == '$') {
	    if (p[1] == '$') {
		ADDCHAR('$');
		p++; /* step over the $ */
	    } else if (p[1] == BRA) {
		/* variable name */
		register char *End;

		p++; /* step over the $ */
		p++; /* step over the BRA */
		for (End = p; *End; End++) {
		    if (*End == KET) {
			char *TheName;
			char *TheValue;

			TheName = emalloc("Variable Name", End - p + 1);
			(void) strncpy(TheName, p, End - p + 1);
			TheName[End - p] = '\0';
			TheValue = GetVariableValue(TheName,
			    MatchFileFormat, MatchFormat,
			    FileNumber, MatchNumber, MatchWithinFile,
			    FID, DocName, FileName, LastDocName,
			    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
			    TextBefore, MatchedText, TextAfter,
			    StartByte, EndByte
			);

			if (TheValue) {
			    ADDSTRING(TheValue);
			} else {
			    Error(E_FATAL,
				"Unknown variable ${%s} in format %s",
				TheName,
				Format
			    );
			}

			p = End;
			break;
		    }
		}

		if (!*End) {
		    Error(E_FATAL,
			"Variable name unterminated (no %c) in format \"%s\"",
			KET,
			Format
		    );
		}
	    } else if (p[1] == '[') {
		/* variable name */
		register char *End;

		p++; /* step over the $ */
		p++; /* step over the [ */
		for (End = p; *End; End++) {
		    if (*End == ']') {
			char *TheName;
			char *TheValue;
			char *TheFormat;

			TheName = emalloc("Variable Name", End - p + 1);
			(void) strncpy(TheName, p, End - p + 1);
			TheName[End - p] = '\0';
			/* Find if there's a format here or not.
			 * The format is indicated by a / as for an
			 * ordinary variable, but unfortunately that's
			 * really hard to tell, because we have to
			 * detect recursion, e.g. in
			 * $[${TextBefore/5.5l}/12r]
			 * or even
			 * $[xxx$[yyy/5r]/5l]
			 *
			 * For now, the rule is that a format can't include
			 * any variables, so we only look back as far as
			 * a / or KET.  But I'd actually really like to
			 * do ${TextBefore/${LeftGapWidth}}.
			 * Sigh.
			 */
			TheFormat = &TheName[End - p];
			while (--TheFormat > TheName) {
			    if (*TheFormat == KET) {
				TheFormat = "";
				break;
			    } else if (*TheFormat == '/') {
				*TheFormat = '\0';
				++TheFormat;
				break;
			    }
			}

			if (TheFormat == TheName) {
			    TheFormat = "";
			}

			TheValue = sPrintWithFormat(TheName,
			    MatchFileFormat, MatchFormat,
			    FileNumber, MatchNumber, MatchWithinFile,
			    FID, DocName, FileName, LastDocName,
			    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
			    TextBefore, MatchedText, TextAfter,
			    StartByte, EndByte
			);
			if (TheValue) {
			    char *z = FormatStringValue(
				TheFormat,
				TheName,
				TheValue
			    );
			    ADDSTRING(z);
			    efree(TheValue);
			}
			efree(TheName);

			p = End;
			break;
		    }
		}

		if (!*End) {
		    Error(E_FATAL,
			"format unterminated ($[ with no ]) in format \"%s\"",
			Format
		    );
		}

	    } else { /* $ followed by something else */
		Error(E_FATAL|E_MULTILINE,
		    "$-sign in a format must be followed by $ or %cvariable%c;",
		    BRA, KET
		);
		Error(E_FATAL|E_MULTILINE,
		    "format contained unexpected character \"%c\" in \"%s\"",
		    *p, Format
		);
		Error(E_FATAL|E_MULTILINE|E_LASTLINE,
		    "%s -L gives the list of valid variables.",
		    progname
		);
	    }
	} else { /* not a $ sign */
	    ADDCHAR(*p);
	}
    }
    return Result;
}

OutputMatch(
    MatchFileFormat, MatchFormat,
    MatchNumber, FID, DocName, FileName,
    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
    TextBefore, MatchedText, TextAfter,
    StartByte, EndByte
)
    char *MatchFileFormat;
    char *MatchFormat;
    long MatchNumber;

    t_FID FID;
    char *DocName; /* the name stored in the database */
    char *FileName; /* the expanded file name */
    int NumberOfWordsInPhrase;
    unsigned long BlockInFile;
    int WordInBlock;

    char *TextBefore;
    char *MatchedText;
    char *TextAfter;

    unsigned long StartByte, EndByte;
{
    /* LastDocName is the last document name that was different
     * from the current one.
     * PrevDocName is the documeny name (DocName) from the
     * immediately preceding call to this function,
     * used only so that we can tell when DocName changes.
     */
    static char *LastDocName = 0;
    static char *PrevDocName = 0;
    static int FileNumber = 0;
    static int MatchWithinFile = 0;

    /* see if the filename has changed: */
    if (!PrevDocName || !STREQ(PrevDocName, DocName)) {
	++FileNumber;
	if (LastDocName) {
	    efree(LastDocName);
	}
	LastDocName = PrevDocName;
	PrevDocName = emalloc("LastDocName", strlen(DocName) + 1);
	(void) strcpy(PrevDocName, DocName);

	MatchWithinFile = 1;

	if (RangeOfFilesToPrint) {
	    if (!WantedInRange(FileNumber, RangeOfFilesToPrint)) {
		if (LargerThanRangeTop(FileNumber, RangeOfFilesToPrint)) {
		    exit(0);
		}
		return;
	    }
	}

	if (MatchFileFormat) {
	    char *p = sPrintWithFormat(MatchFileFormat,
		MatchFileFormat, MatchFormat,
		FileNumber, MatchNumber, MatchWithinFile,
		FID, DocName, FileName, LastDocName, 
		NumberOfWordsInPhrase, BlockInFile, WordInBlock,
		TextBefore, MatchedText, TextAfter, StartByte, EndByte
	    );
	    fputs(p, stdout);
	    efree(p);
	}
    } else {
	++MatchWithinFile;
    }

    if (RangeOfMatchesToPrint) {
	if (!WantedInRange(MatchNumber, RangeOfMatchesToPrint)) {
	    if (LargerThanRangeTop(MatchNumber, RangeOfMatchesToPrint)) {
		exit(0);
	    }
	    return;
	}
    }

    if (MatchFormat) {
	char *p = sPrintWithFormat(MatchFormat,
	    MatchFileFormat, MatchFormat,
	    FileNumber, MatchNumber, MatchWithinFile,
	    FID, DocName, FileName, LastDocName,
	    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
	    TextBefore, MatchedText, TextAfter, StartByte, EndByte
	);
	fputs(p, stdout);
	efree(p);
    }
}

#define IsDirectionChar(ch) ((ch) == 'l' || (ch) == 'r' || (ch) == 'c')

static char *
FormatStringValue(Format, VariableName, Value)
    char *Format;
    char *Value;
    char *VariableName;
{
    static char *Result = 0;
    static unsigned int ResultLength = 0;

    int MinimumWidth = 0;
    int MaximumWidth = 0;
    int LeftPad;
    int RightPad;
    int ValueLength;
    register char *p;
    char PadDirection = 'l';
    char TruncateDirection = 'r';
    int CharsToPrint; /* No. of characters to take from the value string */

    /* Format syntax:
     * minwidth [maxwidth] [l|r|c]
     */
    
    if (Value && *Value) {
	CharsToPrint = ValueLength = strlen(Value);
    } else {
	CharsToPrint = ValueLength = 0;
    }

    for (p = Format; isspace(*p); p++)
	;
    
    if (!*p) {
	/* no format, or empty format, return the value unchanged: */
	return Value ? Value : "";
    } else if (*p == '*') {
	MinimumWidth = 0;
	p++;
    } else if (!isdigit(*p)) {
	Error(E_FATAL,
	    "${%s/%s}: format must start with * or minimum width, not \"%c\"",
	    VariableName, Format, *p
	);
    } else { /* a digit */
	do {
	    MinimumWidth *= 10;
	    MinimumWidth += *p++ - '0';
	} while (*p && isdigit(*p));

	if (IsDirectionChar(*p)) {
	    PadDirection = (*p);
	    p++;
	} else if (*p && !isspace(*p)) {
	    Error(E_FATAL,
		"${%s/%s}: unexpected format character %c, use l, c, r or none",
		VariableName, Format
	    );
	}
    }

    while (isspace(*p)) {
	p++;
    }

    MaximumWidth = MinimumWidth;
    LeftPad = RightPad = 0;

    if (*p == '*') {
	/* ignored */
    } else if (isdigit(*p)) {
	MaximumWidth = 0;
	do {
	    MaximumWidth *= 10;
	    MaximumWidth += *p++ - '0';
	} while (isdigit(*p));

	if (IsDirectionChar(*p)) {
	    TruncateDirection = (*p);
	    p++;
	} else if (*p && !isspace(*p)) {
	    Error(E_FATAL,
		"${%s/%s}: unexpected format character %c, use l, c, r or none",
		VariableName, Format
	    );
	}

	if (CharsToPrint > MaximumWidth) {
	    /* need to truncate the value */
	    CharsToPrint = MaximumWidth;
	}
    }

    if (MinimumWidth > CharsToPrint) {
	switch (PadDirection) {
	case 'l': /* left justify, i.e. pad at the right */
	    RightPad = MinimumWidth - CharsToPrint;
	    break;
	case 'c':
	    LeftPad = (MinimumWidth - CharsToPrint + 1) / 2;
	    RightPad = (MinimumWidth - CharsToPrint) / 2;
	    break;
	case 'r': /* right justify, i.e. pad at the left */
	    LeftPad = MinimumWidth - CharsToPrint;
	    break;
	}
    }

    if (!Result) {
	ResultLength = LeftPad + CharsToPrint + RightPad + 1;
	Result = emalloc("FormatStringValueResult", ResultLength);
    } else if (ResultLength < LeftPad + CharsToPrint + RightPad + 1) {
	ResultLength = LeftPad + CharsToPrint + RightPad + 1;
	Result = erealloc(Result, ResultLength);
    }

    p = Result;

    while (LeftPad-- > 0) {
	*p++ = ' ';
    }

    /* put out the string */
    if (Value && *Value) {
	register char *q;

	switch (TruncateDirection) {
	case 'r':
	default:
	    /* chop off the right hand end */
	    for (q = Value; q - Value < CharsToPrint; q++) {
		*p++ = (*q);
	    }
	    break;
	case 'l':
	    /* chop off the left hand end */
	    {
		register int i = 0;
		for (
		    q = &Value[ValueLength - CharsToPrint];
		    i < CharsToPrint;
		    q++, i++
		) {
		    *p++ = (*q);
		}
	    }
	    break;
	case 'c':
	    {
		register int i = 0;

		for (
		    q = &Value[(ValueLength - CharsToPrint)/2];
		    i < CharsToPrint;
		    q++, i++
		) {
		    *p++ = (*q);
		}
	    }
	    break;
	}
    }

    while (RightPad-- > 0) {
	*p++ = ' ';
    }

    *p = '\0';

    return Result; /* nothing left to print */
}

/* global variables (!) required for termcap: */
int ospeed = 9600; /* actually don't care about the value really */
int PC = '\0';

#define TI_STRING	01
#define TI_NUMBER	02
#define TI_BOOLEAN	03 /* NOTDONE */

static char *tgetStringPointer;
static char *tgetString;
static unsigned int tgetStringLength = 0;

static int
AddToString(ch)
    int ch;
{
    if (tgetStringPointer - tgetString >= tgetStringLength + 2) {
	int Position = tgetStringPointer - tgetString;

	tgetStringLength += 20;
	tgetString = erealloc(tgetString, tgetStringLength);
	tgetStringPointer = &tgetString[Position];
    }
    *tgetStringPointer = ch;
    *++tgetStringPointer = '\0';
}

char *
GetTermInfoString(Name, ValueType)
    char *Name;
    int ValueType;
{
    static char *TerminalEntry = 0;
    char *Result;

    extern char *tgetstr();

    static char Buffer[1024];
    char *p = Buffer;

    if (!TerminalEntry) { /* initialise */
	char *TerminalName = getenv("TERM");
	int Status;

	TerminalEntry = emalloc("Terminal Entry", 1024); /* see termcap(3X) */

	if (!TerminalName) {
	    TerminalName = "unknown";
	}

#ifdef CURSESX
	setupterm(TerminalEntry, fileno(stdout), &Status);
#else
	Status = tgetent(TerminalEntry, TerminalName);
#endif
	switch (Status) {
	case 0:
#ifdef CURSESX
	    Error(E_FATAL,
		"Terminal type \"%s\" has no terminfo entry",
		TerminalName
	    );
#else
	    Error(E_FATAL,
		"Terminal type \"%s\" has no termcap entry",
		TerminalName
	    );
#endif
	case 1: /* OK */
	    break;
	case -1:
	default:
	    Error(E_FATAL|E_SYS,
		"couldn't access termcap file or terminfo database for \"%s\"",
		TerminalName
	    );
	}
    }

    /* now we have a terminal entry */

    switch (ValueType) {
    case TI_STRING:
	p = &Buffer[0];
	Result = tgetstr(Name, &p);
	break;
    case TI_NUMBER:
	{
	    int n = tgetnum(Name);

	    (void) sprintf(Buffer, "%d", Name);
	    Result = &Buffer[0];
	    p = &Result[1];
	}
	break;
    case TI_BOOLEAN:
	{
	    int val = tgetflag(Name);

	    (void) sprintf(Buffer, "%d", Name);
	    Result = &Buffer[0];
	    p = &Result[1];
	}
    default:
	Error(E_BUG|E_FATAL,
	    "GetTermInfoString(%s, %d) where %d is not a supported type",
	    Name, ValueType, ValueType
	);
    }

    if (AsciiTrace > 1) {
	if (!Result || (p == Buffer)) {
#ifdef CURSESX
	    (void) sprintf(Buffer,
		"[terminfo capability \"%s\" not found]",
		Name
	    );
#else
	    (void) sprintf(Buffer,
		"[field \"%s\" not found in termcap entry]",
		Name
	    );
#endif
	    Result = Buffer;
	}
    }

    if (!tgetStringLength) {
	tgetStringLength = strlen(Result) + 10; /** allow for padding **/
	tgetString = emalloc("tgetString", tgetStringLength);
    }
    tgetStringPointer = tgetString;
    *tgetString = '\0';

    (void) tputs(Result, 1, AddToString);
    return tgetString;
}

static char *
GetVariableValue(VariableName,
    MatchFileFormat, MatchFormat,
    FileNumber, MatchNumber, MatchWithinFile,
    FID, DocName, FileName, LastDocName,
    NumberOfWordsInPhrase, BlockInFile, WordInBlock,
    TextBefore, MatchedText, TextAfter, StartByte, EndByte
)
    char *VariableName;
    char *MatchFileFormat;
    unsigned long FileNumber;
    long MatchNumber;
    unsigned long MatchWithinFile;
    char *MatchFormat;
    t_FID FID;
    char *DocName; /* the name stored in the database */
    char *FileName; /* the expanded file name */
    char *LastDocName; /* the previous DocName */
    int NumberOfWordsInPhrase;
    unsigned long BlockInFile;
    int WordInBlock;
    char *TextBefore;
    char *MatchedText;
    char *TextAfter;
    unsigned long StartByte, EndByte;
{
    char *DownCasedName;
    char *Format = "";
    register char *p;

    if (!VariableName) {
	Error(E_FATAL|E_BUG,
	    "GetVariableValue:%s:%d: VariableName is a null pointer",
	    __FILE__, __LINE__, VariableName
	);
    }

    for (p = VariableName; *p; p++) {
	if (*p == '/') {
	    *p = '\0';
	    Format = ++p;
	}
    }

    if (!*VariableName) {
	Error(E_FATAL|E_BUG,
	    "GetVariableValue:%s:%d: VariableName is an empty string",
	    __FILE__, __LINE__, VariableName
	);
    }

    DownCasedName = DownCase(VariableName);
    /* DownCase uses a static buffer, no need to free the result */

    /* the variables are sorted in approximate expected order of frequency of
     * use within each case:
     */

    switch (DownCasedName[0]) {
    case 'b':
	if (STREQ(DownCasedName, "blockinfile")) {
	    return NumberToString(Format, VariableName, BlockInFile, "%lu");
	}
	if (STREQ(DownCasedName, "backspace") || STREQ(DownCasedName, "bs")) {
	    return FormatStringValue(Format, VariableName, "\b");
	}
	break;
    case 'c':
	if (STREQ(DownCasedName, "cr") || /* see also return */
			    STREQ(DownCasedName, "carriagereturn")) {
	    return FormatStringValue(Format, VariableName, "\r");
	}
	break;
    case 'd':
	if (STREQ(DownCasedName, "docname")) {
	    return FormatStringValue(Format, VariableName, DocName);
	}
	if (STREQ(DownCasedName, "doctitle")) {
	    extern char *GetDocumentTitle();
	    char *TheTitle = GetDocumentTitle(FID, DocName);
	    return FormatStringValue(Format, VariableName, TheTitle);
	}
	break;
    case 'e':
	if (STREQ(DownCasedName, "endbyte")) {
	    return NumberToString(Format, VariableName, EndByte, "%lu");
	}
	if (STREQ(DownCasedName, "escape") || STREQ(DownCasedName, "esc")) {
	    return FormatStringValue(Format, VariableName, "\027");
	}
	break;
    case 'f':
	if (STREQ(DownCasedName, "filename")) {
	    return FormatStringValue(Format, VariableName, FileName);
	}
	if (STREQ(DownCasedName, "filenumber")) {
	    return NumberToString(Format, VariableName, FileNumber, "%lu");
	}
	if (STREQ(DownCasedName, "fid")) {
	    return NumberToString(Format, VariableName, FID, "%lu");
	}
	if (STREQ(DownCasedName, "ff") || STREQ(DownCasedName, "formfeed")) {
	    return FormatStringValue(Format, "\f", MatchFormat);
	}
	break;
    case 'g':
	if (STREQ(DownCasedName, "gap")) {
	    static char *Gap = 0;

	    if (!Gap) {
		register char *p;

		Gap = emalloc("Gap String", GapWidth + 1);
		for (p = Gap; p - Gap < GapWidth; p++) {
		    *p = ' ';
		}
		*p = '\0';
	    }
	    return FormatStringValue(Format, "Gap", Gap);
	}
	break;
    case 'l':
	if (STREQ(DownCasedName, "leftpad")) {
	    static char *LeftPad = 0;
	    int Length;
	    register char *p;

	    if (LeftPad) {
		(void) efree(LeftPad);
	    }

	    Length = LeftWidth - strlen(TextBefore);
	    if (Length <= 0) return "";
	    LeftPad = emalloc("LeftPad buffer", Length + 1);
	    for (p = LeftPad; *p; p++) {
		*p = ' ';
	    }
	    *p = '\0';
	    return FormatStringValue(Format, VariableName, LeftPad);
	}
	if (STREQ(DownCasedName, "lastdocname")) {
	    return FormatStringValue(Format, VariableName, LastDocName);
	}
	if (STREQ(DownCasedName, "lf") || STREQ(DownCasedName, "linefeed")) {
	    return FormatStringValue(Format, "\n", MatchFormat);
	}
	break;
    case 'm':
	if (STREQ(DownCasedName, "matchedtext")) {
	    return FormatStringValue(Format, VariableName, MatchedText);
	}
	if (STREQ(DownCasedName, "matchnumber")) {
	    return NumberToString(Format, VariableName, MatchNumber, "%lu");
	}
	if (STREQ(DownCasedName, "matchwithinfile")) {
	    return NumberToString(Format, VariableName, MatchWithinFile, "%lu");
	}
	if (STREQ(DownCasedName, "matchlength")) {
	    return NumberToString(
		Format,
		VariableName,
		EndByte - StartByte,
		"%lu"
	    );
	}
	if (STREQ(DownCasedName, "matchfileformat")) {
	    return FormatStringValue(Format, VariableName,  MatchFileFormat);
	}
	if (STREQ(DownCasedName, "matchformat")) {
	    return FormatStringValue(Format, VariableName, MatchFormat);
	}
	break;
    case 'n':
	if (STREQ(DownCasedName, "numberofwordsinphrase")) {
	    return NumberToString(Format,
		VariableName,
		NumberOfWordsInPhrase,
		"%d"
	    );
	}
	if (STREQ(DownCasedName, "nl") || STREQ(DownCasedName, "newline")) {
	    return FormatStringValue(Format, VariableName, "\n");
	}
	break;
    case 'q':
	if (STREQ(DownCasedName, "quote")) { /* see also cr */
	    return FormatStringValue(Format, VariableName, "\"");
	}
	break;
    case 'r':
	if (STREQ(DownCasedName, "return")) { /* see also cr */
	    return FormatStringValue(Format, VariableName, "\r");
	}
	if (STREQ(DownCasedName, "revision")) { /* see also cr */
	    return FormatStringValue(Format, VariableName, Revision);
	}
	break;
    case 's':
	if (STREQ(DownCasedName, "startbyte")) {
	    return NumberToString(Format, VariableName, StartByte, "%lu");
	}
	break;
    case 't':
	if (STREQ(DownCasedName, "textbefore")) {
	    return FormatStringValue(Format, VariableName, TextBefore);
	}
	if (STREQ(DownCasedName, "textafter")) {
	    return FormatStringValue(Format, VariableName, TextAfter);
	}
	if (DownCasedName[1] == '.') { /* t.terminfo-name */
	    return FormatStringValue(
		Format,
		VariableName,
		GetTermInfoString(&VariableName[2], TI_STRING)
	    );
	}
	if (DownCasedName[1] == '#') { /* t#terminfo-name-number */
	    return FormatStringValue(
		Format,
		VariableName,
		GetTermInfoString(&VariableName[2], TI_NUMBER)
	    );
	}
	if (DownCasedName[1] == '?') { /* t?terminfo-boolean */
	    return FormatStringValue(
		Format,
		VariableName,
		GetTermInfoString(&VariableName[2], TI_BOOLEAN)
	    );
	}
	break;
    case 'v':
	if (STREQ(DownCasedName, "variablename" )) {
	    return FormatStringValue(Format, VariableName, VariableName);
	}
	break;
    case 'w':
	if (STREQ(DownCasedName, "wordinblock")) {
	    return NumberToString(Format, VariableName, WordInBlock, "%d");
	}
	break;
    }
    if (AsciiTrace) {
	Error(E_WARN,
	    "GetVariableValue(%s) -> \"%s\" drew a blank, sorry.",
	    VariableName,
	    DownCasedName
	);
    }
    return (char *) 0; /* unknown variable */
}

static char *VariableHelp[] =
{
    "Variables recognised with -s and -S options:",
    "",
    "All variable names are converted to lower case beforew being used,",
    "so you can type them how you want.  Exeption: variables with names",
    "starting with \"t.\" are retained as they were given.",
    "",
    "[1] ASCII Character names:",
    "BS, CR FF, LF, NL, BackSpace, ESC, Escape, FormFeed, LineFeed, NewLine,",
    "Return, Quote (\"), CarriageReturn -- each replaced by the",
    "corresponsding ASCII character.",
    "",
    "[2] Files and Documents",
    "DocName -- the name of the current document, as stored in the database",
    "FileName -- the apsolute path corresponding to ${DocName}",
    "DocTitle -- the title of the document",
    "FID -- the File Identifier Number of the document",
    "FileNumber -- starts at 1, increases for each new document in the output",
    "LastDocName -- the last value of DocName that was different; the format",
    "  given with -S (${MatchFileFormat}) is used each time DocName changes.",
    "",
    "[3] Match-Specific Variables",
    "BlockInFile, WordInBlock -- these determine the location of the match",
    "NumberOfWordsInPhrase -- the length in words of the phrase matched",
    "TextBefore -- the text in the document immediately before the match",
    "MatchedText -- the document text that exactly matches the phrase",
    "TextAfter -- the text in the document immediately after the match",
    "MatchNumber -- starts at 1 and increases for each match; the format",
    "  given with -s (${MatchFormat}) is used to print each match.",
    "MatchWithinFile -- like MatchNumber but reset for each new document",
    "StartByte -- the byte offset in the file at which the match begins",
    "EndByte -- the byte offset in the file at which the match ends",
    "Matchlength -- length in bytes of ${MatchedText} (EndByte - StartByte)",
    "",
    "[4] Terminal-Specific Variables",
    "t.so, t.se -- start and end stand-out mode (usually reverse video)",
    "t.xx -- xx is looked up in the termcap/terminfo database for your",
    "terminal; xx can be ay string.  If this doesn't do what you want,",
    "using `tput xx` is often better (/usr/5bin/tput xx for SunOS 4.x).",
    "",
    "[5] Output-Formatting Variables",
    "Gap -- a string of spaces, corresponding to the -g command-line argument.",
    "LeftPad -- enough spaces to pad ${TextBefore} to the width given in -l n",
    "RightPad -- enough spaces to pad ${TextAfter} to the width given in -r n",
    "MatchFileFormat - the argument given to -S, or the default value",
    "Matchformat - the argument given to -s, or the default value",
    "",
    "Use these variables by including (for example) ${TextBefore} in the",
    "argument to the -s or -S command-line options.",
    "",
    "You can also give a width specification, e.g. ${TextBefore/9l 12r};",
    "this means that the text will be padded with spaces on the right if",
    "it is shorter then 9 characters, making it left-justified; if it is",
    "longer than 12 characters, it will be truncated from the left, so that",
    "the right-hand end is retained.",
    "Valid characters are l, r and c; a width of * means to use the actual",
    "length of the string.",
    (char *) 0
};

static void
ListVariableValues()
{
    puts("");
    puts("Current settings of variables with default values:");
    printf("  ${MatchFileFormat}=\"%s\"\n", MatchFileFormat);
    printf("  ${MatchFormat}=\"%s\"\n", MatchFormat);
    printf("  ${Gap}=\"");;
    {
	int i;
	for (i = 0; i < GapWidth; i++) {
	    putchar(' ');
	}
    }
    printf("\"\n");
    printf("  ${Revision}=\"%s\"\n", Revision);
}

static void
ListVariables()
{
    char **p;

    fflush(stderr); /* in case there were error messages */
    for (p = VariableHelp; *p; p++) {
	printf("  %s\n", *p);
    }
    ListVariableValues();
    fflush(stdout); /* so it doesn't get mingled with more error messages */
}
